/***************************************************************************
   plugin.cpp  -  plugin API
                             -------------------
    begin                : Wed Jul 03 2002
    copyright            : (C) 2002 by red0x
    email                : red0x@users.sourceforge.net
    rcsid                : $Id: plugin.c,v 1.3 2002/08/08 08:20:54 red0x Exp $
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/** @file plugin.c
  * File contains plugin API Implementation\n
  * Plugin API version 1.0 \n
  * To make a plugin, look in plugin.h for the plugin_t structure,
  * and make sure your plugins can provide all of the needed methods.
  * If you dont need a method, for now, just make it return PLUGIN_OK
  * Until I get a more detailed API documentation, that is how we will
  * do it.
  * @test Input Plugins
  * @test Output Plugins
  */

#include "plugin.h"
#include "util.h"
#include <assert.h>
#include <dirent.h>
#include <dlfcn.h>
#include <stdlib.h>

static const char plugid[] =
        "\100$ autosec/sonar: $Id: plugin.c,v 1.3 2002/08/08 08:20:54 red0x Exp $";

/// list of input plugins
plugin_list_t input_list[MAX_INP];      // INPUT PLUGINS  -- file, range, random, etc.
/// list of output plugins
plugin_list_t output_list[MAX_OUT];     // OUTPUT PLUGINS -- human, machine, etc..
/// list of network plugins
plugin_list_t network_list[MAX_NET];    // NETWORK PLUGINS -- icmp, tcp, snmp, etc..

/// plugin id system PID
int first_pid;
/// Number of input plugins
unsigned long int num_inp,
        /// number of output plugins
  num_out,
        /// number of network plugins
  num_net;

/** @name Plugin API Internals
  * These are behind the scenes functions that your plugins, 
  * nor you, should ever need to call.
  * Do not call these functions!
 */
//@{

/** @fn plugin_ready()
  * Plugin API Interal -- DO NOT CALL FROM PLUGINS\n
  * Checks if the plugin subsystem is ready
  * @return 1 if the system is ready
  */
int inline plugin_ready()
{
        if(plugid)
        {
        }                       // useless reference to avoid warnings
        // are they all defined?
        if(!input_list || !output_list || !network_list)
                return 0;
        return 1;
}

/** @fn plugin_init(plugin_list_t * x)
  * Plugin API Interal -- DO NOT CALL FROM PLUGINS\n
  * Sets the plugin_list_t member "plug" to NULL
  * @param plugin_list plugin_list_t pointer
  */
void inline plugin_init(plugin_list_t * plugin_list)
{
        // set the plugin to NULL
        plugin_list->plug = NULL;
}

/** @fn plugin_newid()
  * Plugin API Interal -- DO NOT CALL FROM PLUGINS\n
  * Gets a new pluginID for your plugin, once loaded
  * @return A new, unique plugin ID
  * @sideeffect first_pid is changed to the new plugin ID number
  */
int inline plugin_newid()
{
        return (unsigned long int) (first_pid += ((unsigned int)
                                                  (200.0 * rand() /
                                                   (RAND_MAX + 64.0))));
}

//@}

/** @name Plugin API -- Sonar side
  * These are behind the scenes functions that sonar uses
  * to load and initialize your plugins.  You should never
  * need access to these functions.
  * Do not call these functions!
 */
//@{

/**
  * @fn plugin_start(sonar_t *mthis)
  * Startup and load the plugin API
  * @param mthis Pointer to a filled sonar_t structure with all of sonar's
  * pertinent function pointers and information
  * @return PLUGIN_OK if ok, PLUGIN_ERROR otherwise
  * @sideeffect first_pid, num_inp, num_out, num_net are all returned to 0
  */
int plugin_start(sonar_t * mthis)
{
        //input_list = plugin_new();
        if(!input_list)
                return PLUGIN_ERROR;
        plugin_init(&input_list[0]);
        //output_list = plugin_new();
        if(!output_list)
                return PLUGIN_ERROR;
        plugin_init(&output_list[0]);
        //network_list = plugin_new();
        if(!network_list)
                return PLUGIN_ERROR;
        plugin_init(&network_list[0]);
        first_pid = 0;
        num_inp = num_out = num_net = 0;
        return PLUGIN_OK;
}

/**
  * @fn plugin_end(void)
  * Unload the plugin API
  * @return PLUGIN_OK
  * @sideeffect first_pid, num_inp, num_out, and num_net are returned to 0
  */
int plugin_end(void)
{
        //delete [] input_list;
        //delete [] output_list;
        //delete [] network_list;
        first_pid = 0;
        num_inp = num_out = num_net = 0;
        return PLUGIN_OK;
}

/**
  * @fn plugin_loadall(sonar_t *mthis)
  * Find and load all plugins in the plugin directory
  * @param mthis Pointer to a filled sonar_t structure
  * @return PLUGIN_OK if ok, PLUGIN_ERROR otherwise
  * @test Loading input plugins, loading output plugins
  */
int plugin_loadall(sonar_t * mthis)
{
        // directory to search defined as PLUGINDIR
        DIR *dir;
        void *plugin;
        void *(*plugin_init) (void *);
        dir = opendir(PLUGINDIR);
        if(dir)
        {
                struct dirent *entry;
                while((entry = readdir(dir)) != NULL)
                {
                        char str[1024];
                        int len = strlen(entry->d_name);
                        if((entry->d_name[len - 3] == '.') &&
                           (entry->d_name[len - 2] == 's') &&
                           (entry->d_name[len - 1] == 'o'))
                        {
                                //found it
                                memset(str, 0, 1024);
                                snprintf(str, 1023, "%s/%s", PLUGINDIR,
                                         entry->d_name);
                                if((plugin = dlopen(str, RTLD_LAZY)) == NULL)
                                {
                                        char *dl_error_msg =
                                                (char *) dlerror();
                                        sonar_msg("could not load %s:\n%s\n",
                                                  str, dl_error_msg);
                                        continue;
                                }

                                plugin_init =
                                        (void *(*)(void *)) dlsym(plugin,
                                                                  "plugin_init");
                                if(plugin_init != NULL)
                                {
                                        plugin_t *plug;
                                        plug = (plugin_t *)
                                                plugin_init(mthis);
                                        if(!plug)
                                        {
                                                sonar_msg
                                                        ("Could not load %s:\nplugin_init failed\n",
                                                         str);
                                        }
                                        else
                                                plugin_add(plug->type, plug,
                                                           mthis, plugin);
                                }
                                else
                                {
                                        char *dl_error_msg =
                                                (char *) dlerror();
                                        sonar_msg
                                                ("could not load plugin_init in %s:\n%s\n",
                                                 str, dl_error_msg);
                                }
                        }
                }
        }
        closedir(dir);
        return PLUGIN_OK;
}

/**
  * @fn plugin_killall(void)
  * Unload all plugins
  * @return PLUGIN_OK if ok, PLUGIN_ERROR otherwise
  * @sideeffect num_inp, num_out, and num_net returned to 0
  * @sideeffect plug->plugin_destroy is asserted not null
  */
int plugin_killall(void)
{
        int err = PLUGIN_OK;
        //get all input, output, network plugins
        //call the plugin_destroy function
        for(; num_inp > 0; num_inp--)
        {
                int my_err;
                assert(input_list[num_inp - 1].plug != NULL);
                assert(input_list[num_inp - 1].plug->plugin_destroy != NULL);
                //if plugin is also an output plugin, let that call its destroy function
                if(!(input_list[num_inp - 1].plug->type & OUTPUT_PLUGIN))
                {
                        if((my_err =
                            input_list[num_inp - 1].plug->plugin_destroy()) !=
                           PLUGIN_UNLOADED)
                                err = my_err;
                        input_list[num_inp - 1].plug = NULL;
			dlclose(input_list[num_inp - 1].handle);
                }
        }
        for(; num_out > 0; num_out--)
        {
                int my_err;
                assert(output_list[num_out - 1].plug != NULL);
                assert(output_list[num_out - 1].plug->plugin_destroy != NULL);
                //if plugin is also a network plugin, let that call it destroy function
                if(!(output_list[num_out - 1].plug->type & NETWORK_PLUGIN))
                {
                        if((my_err =
                            output_list[num_out -
                                        1].plug->plugin_destroy()) !=
                           PLUGIN_UNLOADED)
                                err = my_err;
                        output_list[num_out - 1].plug = NULL;
			dlclose(output_list[num_out - 1].handle);
                }
        }
        for(; num_net > 0; num_net--)
        {
                int my_err;
                assert(network_list[num_net - 1].plug != NULL);
                assert(network_list[num_net - 1].plug->plugin_destroy !=
                       NULL);
                if((my_err =
                    network_list[num_net - 1].plug->plugin_destroy()) !=
                   PLUGIN_UNLOADED)
                        err = my_err;
                network_list[num_net - 1].plug = NULL;
		dlclose(network_list[num_net - 1].handle);
        }
        return err;
}

/**
  * @fn plugin_add(int type, plugin_t *p_data, sonar_t *mthis, void *hptr)
  * Add plugin to proper plugin_list
  * @param type Plugin type (ex, for input and output, it would be INPUT_PLUGIN | OUTPUT_PLUGIN
  * @param p_data Pointer to filled plugin_t structure
  * @param mthis Pointer to a filled sonar_t structure
  * @param hprt Handle returned by dlopen(3)
  * @return PLUGIN_OK if ok, PLUGIN_ERROR otherwise
  */
int plugin_add(int type, plugin_t * p_data, sonar_t * mthis, void *hptr)
{
        if(!plugin_ready())
        {
                sonar_msg("plugin_add: plugins not ready\n");
                return PLUGIN_ERROR;
        }
        {
                assert(p_data != NULL);
                assert(mthis != NULL);
        }
        if(type & INPUT_PLUGIN)
        {
                plugin_init(&(input_list[num_inp]));

                input_list[num_inp].plug = p_data;
		input_list[num_inp].handle = hptr;
                input_list[num_inp].plug->pluginID = plugin_newid();
                num_inp++;
                if(num_inp > MAX_INP)
                {
                        DEBUG("TOO MANY INPUT PLUGINS, ABORTING!\n");
                        exit(-1);
                }
        }
        if(type & OUTPUT_PLUGIN)
        {
                plugin_init(&(output_list[num_out]));

                output_list[num_out].plug = p_data;
		output_list[num_out].handle = hptr;
                output_list[num_out].plug->pluginID = plugin_newid();
                num_out++;
                if(num_out > MAX_OUT)
                {
                        DEBUG("TOO MANY OUTPUT PLUGINS, ABORTING...\n");
                        exit(-1);
                }
        }
        if(type & NETWORK_PLUGIN)
        {
                plugin_init(&(network_list[num_net]));

                network_list[num_net].plug = p_data;
		network_list[num_net].handle = hptr;
                network_list[num_net].plug->pluginID = plugin_newid();
                num_net++;
                if(num_net > MAX_NET)
                {
                        DEBUG("TOO MANY NETWORK PLUGINS, ABORTING...\n");
                        exit(-1);
                }
        }
        return PLUGIN_OK;
}

//@}

/** @name Plugin API -- Hooks
  * These are hook functions used to print usage and status information
  * as well as handling command line options
  * Do not call these functiions!
 */
//@{

/**
  * @fn plugin_status_hook(void)
  * Call each plugin's status function to print extra status reports
  * at the status screen
  */
void plugin_status_hook(void)
{
        unsigned int plug;
        //get all input, output, network plugins
        //call the plugin_status function

        for(plug = 0; plug < num_inp; plug++)
        {
                {
                        assert(input_list[plug].plug != NULL);
                        assert(input_list[plug].plug->plugin_status != NULL);
                }
                // each plugin has access to sonar's output function
                // so dont worry about output
                input_list[plug].plug->plugin_status();
        }
        for(plug = 0; plug < num_out; plug++)
        {
                {
                        assert(output_list[plug].plug != NULL);
                        assert(output_list[plug].plug->plugin_status != NULL);
                }
                // each plugin has access to sonar's output function
                // so dont worry about output
                output_list[plug].plug->plugin_status();
        }
        for(plug = 0; plug < num_net; plug++)
        {
                {
                        assert(network_list[plug].plug != NULL);
                        assert(network_list[plug].plug->plugin_status !=
                               NULL);
                }
                // each plugin has access to sonar's output function
                // so dont worry about output
                network_list[plug].plug->plugin_status();
        }
        return;
}

/**
  * @fn plugin_usage_hook(void)
  * Call each plugin's usage function to print extra usage information
  * at the usage screen
  */
void plugin_usage_hook(void)
{
        unsigned int plug;
        //get all input, output, network plugins
        //call the plugin_usage function

        for(plug = 0; plug < num_inp; plug++)
        {
                {
                        assert(input_list[plug].plug != NULL);
                        assert(input_list[plug].plug->plugin_usage != NULL);
                }
                // each plugin has access to sonar's output function
                // so dont worry about output
                input_list[plug].plug->plugin_usage();
        }
        for(plug = 0; plug < num_out; plug++)
        {
                {
                        assert(output_list[plug].plug != NULL);
                        assert(output_list[plug].plug->plugin_usage != NULL);
                }
                // each plugin has access to sonar's output function
                // so dont worry about output
                output_list[plug].plug->plugin_usage();
        }
        for(plug = 0; plug < num_net; plug++)
        {
                {
                        assert(network_list[plug].plug != NULL);
                        assert(network_list[plug].plug->plugin_usage != NULL);
                }
                // each plugin has access to sonar's output function
                // so dont worry about output
                network_list[plug].plug->plugin_usage();
        }
        return;
}

/**
  * @fn plugin_command_handler(char opt, char *optarg)
  * Call each plugin's plugin_args function to see if it handles this option
  * @param opt Option char as passed from getopt
  * @param optarg Option argument as passed from getopt
  * @return SONAR_ERROR if no plugins handled the argument\n
  * SONAR_OK if someone handled it
  */
int plugin_command_handler(char opt, char *optarg)
{
        //go through the plugins, and see if this option is handled, first network
        // plugins, then input, then output
        // else, return SONAR_ERROR
        unsigned int plug = 0;
        int error = PLUGIN_ERROR;
        for(plug = 0; plug < num_inp; plug++)
        {
                int myerror;
                assert(input_list[plug].plug != NULL);
                assert(input_list[plug].plug->plugin_args != NULL);
                if((myerror =
                    input_list[plug].plug->plugin_args(opt,
                                                       optarg)) !=
                   PLUGIN_ERROR && error == PLUGIN_ERROR)
                        error = myerror;
        }
        for(plug = 0; plug < num_out; plug++)
        {
                int myerror;
                assert(output_list[plug].plug != NULL);
                assert(output_list[plug].plug->plugin_args != NULL);
                if((myerror =
                    output_list[plug].plug->plugin_args(opt,
                                                        optarg)) !=
                   PLUGIN_ERROR && error == PLUGIN_ERROR)
                        error = myerror;
        }
        for(plug = 0; plug < num_net; plug++)
        {
                int myerror;
                assert(network_list[plug].plug != NULL);
                assert(network_list[plug].plug->plugin_args != NULL);
                if((myerror =
                    network_list[plug].plug->plugin_args(opt,
                                                         optarg)) !=
                   PLUGIN_ERROR && error == PLUGIN_ERROR)
                        error = myerror;
        }
        return error;
}

/**
  * @fn plugin_network_command_handler(char *hostname)
  * Call each plugin's plugin_next function to see if it handles this host/ip/option
  * @param hostname Hostname or IP or option from command line
  * @return SONAR_ERROR if no plugins handled the argument\n
  * SONAR_OK if someone handled it
  */
int plugin_network_command_handler(char *hostname)
{
        unsigned int plug;
        int error = PLUGIN_ERROR;
        for(plug = 0; plug < num_net; plug++)
        {
                int myerror;
                assert(network_list[plug].plug != NULL);
                assert(network_list[plug].plug->plugin_next != NULL);
                if((myerror =
                    network_list[plug].plug->plugin_next(hostname)) ==
                   PLUGIN_OK)
                        error = myerror;
        }
        return error;
}

//@}
