/***************************************************************************
   plugin.h  -  plugin API
                             -------------------
    begin                : Wed Jul 03 2002
    copyright            : (C) 2002 by red0x
    email                : red0x@users.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/** @file plugin.h
  * File contains plugin API \n
  * Plugin API 1.0 \n
  * To make a plugin, refer to this file's plugin_t structure
  * and to the example plugins
  */
#ifndef _PLUGIN_H_
#define _PLUGIN_H_

#include "types.h"
#include <string.h>
/// default plugin dir
#define PLUGINDIR  "/usr/local/lib/sonar"

/** @name Error and Success Values
  * return these from your plugin functions
 */
//@{
/// error with last function
#define PLUGIN_ERROR	-1
/// plugin unloaded
#define PLUGIN_UNLOADED -2
/// no response (for network plugins)
#define PLUGIN_NORESP	-3
/// file not found
#define PLUGIN_NOFILE	-4
/// end of file
#define PLUGIN_EOF	-5
/// no permissionsa
#define PLUGIN_PERM	-6
/// not open
#define PLUGIN_NOPEN	-7


/// last function returned ok
#define PLUGIN_OK	1       // must correspond with SONAR_OK
/// plugin loaded
#define PLUGIN_LOADED	2
/// response (for network plugins)
#define PLUGIN_RESP	3

//@}

//#define PLUGIN_NOTREADY -1

/** @name Plugin Types
  * Just '|' these values together for your plugin type.
  * For example for an input/output plugin, do this
  * plug->type = INPUT_PLUGIN | OUTPUT_PLUGIN;
 */
//@{
/// Input plugin
#define INPUT_PLUGIN 	1
/// Output Plugin
#define OUTPUT_PLUGIN	2
/// Network Plugin
#define NETWORK_PLUGIN	4
//@}

/** @name Max Values
  * Maximum values for certain things in the plugin API
 */
//@{

/// Max input plugins
#define MAX_INP		20
/// Max output plugins
#define MAX_OUT		20
/// Max network plugins
#define MAX_NET		20
//@}

/// Sonar side debug script
#define DEBUG(x) sonar_msg("%s:%d: %s\n", __FILE__, __LINE__, x);

/** @struct __plugin_t
  * Definition for the plugin interface
  * This is return to sonar from the plugin_init function
  */
typedef struct __plugin_t
{
        /// unique ID assigned at run time
        unsigned long int pluginID;
        /// plugin type (INPUT_PLUGIN | OUTPUT_PLUGIN for an I/O plugin)
        int type;

        /// last error from this plugin
        int *errno;
        /// ready to use flag, not modified by sonar
        bool ready;

        /// Initialize this plugin
        void *(*plugin_init) (void *in_data);
        /// Destroy this plugin
        int (*plugin_destroy) (void);

        /// Argument handler
        int (*plugin_args) (char opt, char *optarg);

        /// Open the input/output/network source
        int (*plugin_open) (void *c_data, int i_data);
        /// Close the input/output/network source
        int (*plugin_close) (void);
        /// Get next host, returned in hostname, only for input plugins
        /// Set next host, for network plugins
        int (*plugin_next) (char *hostname);

        /// For network plugins
        /// Do the sweep
        int (*plugin_sweep) (void *data);
        /// Keep sweeping? XXX: Scheduled for removal, we wont need this
        int (*plugin_go) (void);

        /// prints, via sonar_t->output(), line(s) to add to sonar_usage()
        void (*plugin_usage) (void);
        /// prints, via sonar_t->output(), line(s) to add to the status report
        void (*plugin_status) (void);

        /// Return the last error
        int (*plugin_last_error) (void);

}plugin_t;

/** @struct __sonar_t
  * Definition for the plugin interface
  * This is passed to the plugin, via plugin_init, from sonar with useful
  * function pointers and information, such as what version
  * of sonar is loading the plugins, and pointers to sonar_msg(char *,...), etc.
  */
typedef struct __sonar_t
{
        /// Major version number
        int ver_major,
        /// Minor version number
		ver_minor,
        /// Revision number
		revision,
        /// Plugin API version
		api;

        /// Pointer to a resolver from sonar
        unsigned long (*resolver) (char *name);
        /// Pointer to a memory copier from sonar
        void (*memput) (char **dest, const char *src, const int len);
        /// Pointer to the Messsage output (ie, with a timestamp) function from sonar
        void (*message) (char *msg, ...);
        /// Pointer to sonar's isip function, to verify ip addresses
        int (*ipverify) (char *ip);

        /// Pointer to sonar's output to console function (no timestamp)
        int (*output) (char *msg, ...);

        /// Pointer to usage function, for error reports
        void (*usage) (const char *);

        /// Pointer to a function to get root priveleges
        int (*getprivs) (void);
        /// Pointer to a function to drop root priveleges
        int (*dropprivs) (uid_t);
}sonar_t;

/** @struct __plugin_list_t
  * Wrapper for plugin_t to be used as an array
  */
typedef struct __plugin_list_t
{
        /// Pointer to a plugin_t structure
        plugin_t *plug;
	/// Handle to library opened with \c dlopen(3)
	void *handle;
}plugin_list_t;

int plugin_start(sonar_t * mthis);
int plugin_end(void);
int plugin_loadall(sonar_t * mthis);
int plugin_killall(void);
int plugin_add(int type, plugin_t * p_data, sonar_t * mthis, void *hptr);
void plugin_status_hook(void);
void plugin_usage_hook(void);
int plugin_command_handler(char opt, char *optarg);
int plugin_network_command_handler(char *hostname);

typedef plugin_t *(*PINIT) (void *);
typedef int (*PDESTROY) (void);
typedef int (*POPEN) (void *, int);
typedef int (*PCLOSE) (void);
typedef int (*PNEXT) (char *);
typedef void (*PHOOK) (void);
typedef int (*PERROR) (void);
typedef int (*PSWEEP) (void *);
typedef int (*POGO) (void);

/** @page PluginAPI Plugin API documentation
    @subsection v2 Version 2
  This page will explain the use and programming of
  sonar plugins.  Sonar has been trimmed down, diverting most
  of the processing to plugins.  These plugins use an API, 
  or Application Programming Interface, which provides 
  for input, output, and network plugins.  That means you can
  create your own plugins to drive different kinds of scans using
  different types of plugins. Version 2 of the API is used in sonar
  1.0BETA1 and above. \n\n
  
  @subsection contents Contents 
   1. \ref api \n
     &nbsp;&nbsp; a. \ref Sonar \n 
     &nbsp;&nbsp;&nbsp;&nbsp; 1. \ref load \n
     &nbsp;&nbsp;&nbsp;&nbsp; 2. \ref args \n
     &nbsp;&nbsp; b. \ref Plugin \n
     &nbsp;&nbsp;&nbsp;&nbsp; 1. \ref pload \n
     &nbsp;&nbsp;&nbsp;&nbsp; 2. \ref pother \n
   2. \ref notes \n
   3. \ref argsnotes \n
   
  @section api The Plugin API 
  \n
  @section Sonar Sonar Internals
  @subsection load Loading plugins
    Sonar will search through \c /usr/local/lib/sonar for all \c *.so 
  files (see \ref plugin_loadall). Sonar calls this function when it is
  ready to load plugins.  It passes \ref plugin_loadall a pointer to a 
  filled in \ref __sonar_t structure.  \ref plugin_loadall searches through
  \c /usr/local/lib/sonar for all \c *.so files.  Any file it finds, it 
  opens with dlopen(3).  If it can't open a file, it reports which file it
  couldn't open and continues looking for more \c *.so files.  \b NOTE: \b sonar
  \b will \b not \b look \b in \b the \b PATH \b or \b the \b working \b directory
  \b for \b plugins. \b It \b will \b only \b look \b in \c /usr/local/lib/sonar.  \n\n
    Once it opens your plugin, it will call dlsym(3) to load your plugin's 
  \ref plugin_init function.  It is imperative that your plugin has a function called
  \ref plugin_init (See \link bare.h bare.h \endlink).  Sonar then calls this function,
  passing it a pointer to a complete \ref __sonar_t structure (note: __sonar_t and 
  sonar_t are interchangeable).  Sonar expects that your plugin_init function return
  a pointer to a completely filled out __plugin_t structure. After this, sonar adds
  a handle and your plugin_t structure to an array of plugins, depending on what
  type of plugin it is.  Sonar will store a handle of your plugin, and will call 
  dlclose when exiting the program. \n\n
  
  @subsection args Argument Handling
    Sonar has a few main arguments, but, with all the plugins, it could have an
  infinite amount of arguments to handle.  To counter the confusion of passing every
  plugin every argument, I made all the plugin arguments have a prefix.  That
  prefix is \c -p.  In other words, anything after -p is passed to \b all of the 
  plugins. For example: processing \c -pd60 will cause sonar to call each plugins'
  plugin_args function (see __plugin_t) with \c d as the opt and \c 60 as the optarg.
  If your plugin has not handled that argument, sonar expects a return value of 
  \ref PLUGIN_ERROR .  If your plugin did handle the argument, sonar expects a return
  value of \ref PLUGIN_OK .  \n\n
   If your plugin is a network plugin, a similar dilemma arises. How can we have more
  than one network plugin, and only run one scan at a time, rather than running them
  all at once.  I decided to go with the prefix solution again.  Each network plugin
  should be designed with a prefix for all hostnames it is supposed to handle.  For 
  example, the \link network_icmp.c network_icmp \endlink plugin's prefix is \c 'I'.
  Your plugin must expect all hostname's to have a prefix, check that the prefix is
  correct, and strip it off.  If it is not correct, sonar expects your plugin to return
  \ref PLUGIN_ERROR to signal that the hostname was not handled by your plugin. \n\n

  @section Plugin Plugin Internals
  @subsection pload Loading

   Once sonar has dlopen(3)'ed your plugin, it calls your plugin_init function.
  Each plugin \b *must* have a function named plugin_init, and this function 
  must return a pointer to a filled int \ref __plugin_t structure.  If your
  plugin is an input/output only plugin, it can leave the plugin_sweep function 
  pointer blank. Your function must also setup its type correctly, or the plugin
  hooks may fail.  To set your plugin's type, set your pthis.type to the correct 
  bitwise ORing ('|') of the different types (See \ref INPUT_PLUGIN, \ref OUTPUT_PLUGIN,
  \ref NETWORK_PLUGIN, and \link bare.c bare.c \endlink).
  Plugins must provide all of the other functions currently in the \ref __plugin_t
  structure to avoid confusion and assert errors.  

  @subsection pother Other Questions
   If you have any questions about how to do anything else with plugins, see 
   \link bare.c bare.c \endlink , \link bare.h bare.h \endlink , and 
   \link plugin.h plugin.h \endlink .  If you want to make a very bare plugin that
   only does a few things, just provide all the functions and have each return 
   either PLUGIN_OK or PLUGIN_ERROR (See \link bare.c bare.c \endlink). 
   Also, feel free to shoot me an email at
   \c red0x@users.sourceforge.net or see me on IRC at \c irc.root.gg:#lameindustries.


  @subsection notes Notes
   Be careful when developing plugins, since
  there is currently no way to avoid overlap of plugin arguments.  If two
  plugins handle the same argument, there is a conflict and things could
  go south.  This situation is current \b NOT detected \n NOR corrected.  \n\n

  After all arguments are parsed, a list of hosts should remain on the 
  command line to be parsed.  If not, currently, sonar bails.  This poses a
  problem for input plugins that will generate a random list of hosts. 
  This problem will be addressed in a later release of the plugin API.
  Once sonar reaches this section, it passes all the "hostnames" on the 
  command line to all the network plugins, seeing if it handles it.
  See \ref plugin_network_command_handler.
  I suggest you declare a prefix in your plugin's usage function to prepend
  to all "hostnames" on the command line, so that only your plugin will handle
  the correct hostnames.  That is the current work around. (see \link rfc793.c 
  rfc793.c \endlink).
  \n\n
  Once all that is taken care of, sonar passes control off to the network
  plugins.  Plugins are currently designed to be run in parallel, so only do
  one pass of the sweep, make your default post-delay 0, and return PLUGIN_OK once
  your network plugin is done sweeping. 
  \n\n
   Plugins handle their arguments by providing a plugin_args function.
  Sonar will call this function with the character following the -p and
  the number or string following that, so you can easily use a getopt
  style switch statement to parse the arguments from there.  If your
  plugin does not handle that particular argument, do not exit, but 
  return \c PLUGIN_ERROR. This will tell sonar that you did not handle it, 
  and give other plugins a chance to handle it.  If you did handle it,
  return \c PLUGIN_OK. (See \ref icmp_args).
  \n\n
  The plugin will, if its a network plugin, receive a list of "hostnames" from
  the command line in its plugin_next function (See \ref icmp_add). If the 
  plugin can handle that "hostname", it should return \c PLUGIN_OK, or 
  \c PLUGIN_ERROR otherwise.
  \n\n
  The plugin's sweep function, if its a network function, should handle
  the sweeps however it wants. I ask that you only perform one sweep at a 
  time, for parallel scanning reasons.
  \n\n
  For more information and examples, see \link network_icmp.c 
  network_icmp.c \endlink and \link network_icmp.h network_icmp.h 
  \endlink

  @subsection argsnotes Notes on Passing Arguments
  Sonar passes all options prefixed with -p to plugins.  So, as a 
  change from previous versions, to change the TTL, you would have to
  type either \c -pt250 or \c -p \c t250. \c -p \c -t \c 250 
  \b WILL \b NOT \b WORK! Make sure the
  plugin option and argument are \b NOT separated by a space, the -p and
  the rest of the option may be separated however. \n
  \n\n
*/

#endif //_PLUGIN_H
